@@ -39,6 +39,8 @@ static mutex_t threads_lists_lock = MUTEX_STATIC_INIT;
 static GList *threads_to_join;
 static GList *threads_running;
 static cond_t threads_cond = COND_STATIC_INIT;
+static mutex_t thread_wakers_lock = MUTEX_STATIC_INIT;
+static GList *thread_wakers;
 
 volatile int rtpe_shutdown;
 
@@ -131,12 +133,30 @@ static gint thread_equal(gconstpointer a, gconstpointer b) {
 	return !pthread_equal(*x, *y);
 }
 
-void threads_join_all(int wait) {
+void threads_join_all(bool cancel) {
 	pthread_t *t;
 	GList *l;
 
-	mutex_lock(&threads_lists_lock);
 	while (1) {
+		if (cancel) {
+			mutex_lock(&thread_wakers_lock);
+			for (l = thread_wakers; l; l = l->next) {
+				struct thread_waker *wk = l->data;
+				mutex_lock(wk->lock);
+				cond_broadcast(wk->cond);
+				mutex_unlock(wk->lock);
+			}
+			mutex_unlock(&thread_wakers_lock);
+
+			mutex_lock(&threads_lists_lock);
+			for (l = threads_running; l; l = l->next) {
+				t = l->data;
+				pthread_cancel(*t);
+			}
+			mutex_unlock(&threads_lists_lock);
+		}
+
+		mutex_lock(&threads_lists_lock);
 		while (threads_to_join) {
 			t = threads_to_join->data;
 			pthread_join(*t, NULL);
@@ -151,19 +171,38 @@ void threads_join_all(int wait) {
 			g_slice_free1(sizeof(*t), t);
 		}
 
-		if (!wait)
-			break;
-		if (!threads_running)
+		if ((!cancel && rtpe_shutdown) || (cancel && !threads_running)) {
+			mutex_unlock(&threads_lists_lock);
 			break;
+		}
 		cond_wait(&threads_cond, &threads_lists_lock);
+		mutex_unlock(&threads_lists_lock);
 	}
-	mutex_unlock(&threads_lists_lock);
+}
+
+void thread_waker_add(struct thread_waker *wk) {
+	mutex_lock(&thread_wakers_lock);
+	thread_wakers = g_list_prepend(thread_wakers, wk);
+	mutex_unlock(&thread_wakers_lock);
+}
+void thread_waker_del(struct thread_waker *wk) {
+	mutex_lock(&thread_wakers_lock);
+	thread_wakers = g_list_remove(thread_wakers, wk);
+	mutex_unlock(&thread_wakers_lock);
+}
+
+static void thread_detach_cleanup(void *dtp) {
+	struct detach_thread *dt = dtp;
+	g_slice_free1(sizeof(*dt), dt);
+	thread_join_me();
 }
 
 static void *thread_detach_func(void *d) {
 	struct detach_thread *dt = d;
 	pthread_t *t;
 
+	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
+
 	t = g_slice_alloc(sizeof(*t));
 	*t = pthread_self();
 	mutex_lock(&threads_lists_lock);
@@ -207,9 +246,10 @@ static void *thread_detach_func(void *d) {
 					dt->priority, strerror(errno));
 	}
 
+	pthread_cleanup_push(thread_detach_cleanup, dt);
 	dt->func(dt->data);
-	g_slice_free1(sizeof(*dt), dt);
-	thread_join_me();
+	pthread_cleanup_pop(true);
+
 	return NULL;
 }
 
